/**********************************************************
* FILE LINKLIST.C version 1.01  - 05/10/94 -
*
* DESCRIPTION:
*  contains the link list's functions.
*
* NOTES:
*
* Copyright 1993 by EZ Software 
*
* Permission to use, copy, modify, and distribute this software and its
* documentation for any purpose and without fee is hereby granted, provided 
* that the above copyright notice appear in all copies and that both that 
* copyright notice and this permission notice appear in supporting 
* documentation, and that the name of EZ Software 
* not be used in advertising or publicity pertaining to distribution of the 
* software without specific written prior permission. 
* EZ Software makes no representations about the 
* suitability of this software for any purpose.  It is provided "as is"
* without express or implied warranty.
*
* EZ Software DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
* SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, 
* IN NO EVENT SHALL EZ Software BE LIABLE FOR ANY SPECIAL,
* INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES RESULTING FROM LOSS OF USE,
* DATA OR PROFITS, WHETHER IN AN ACTION CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
* ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
* SOFTWARE.
*
*
*  Original Author: Laurent Chavey, EZ Software 
*           (chavey@cis.udel.edu)
*            june, 1993
*
* UPDATES: 
*          05/10/94  - fixed bug in add function.
*          01/20/94  - changes search and Search to compare Compare. 
*                    - fixed problem in delete.
*                    - added support for multiple compare function.
*                    - added dump function..
*                    - added support for adding at current rather
*                      than tail.
*                    - added parameter to add function.
*                    - added return standard error code. 
*                    - removed item size from add, added it to main linklist
*                      structure and llnew function.
*                    - Changed Copyright from HANSEN INFORMATION TECHNOLOGY
*                      to EZ Software.
*
**********************************************************/

#define LINKLIST
#include "linklist.h"

/********************* PRIMITIVES ***********************/

/*********************************************************
* FUNCTION: 
*  llnew
*
* PROTOTYPE:
*  linklist *llnew(int, int);
*
* PARAMETERS:
*  int  nCompare  number of compare function used for linklist element.
*  int  nDataSize size of the data stored in linkelement.
*
* RETURN:
*  linklist *  pointer to the new linklist.
*
* DESCRIPTION:
*  crates a link list.
*
* NOTES:
*  compare function should return 1 when found, 0 if not.
*
* UPDATES: 01/20/94  - changes search and Search.
*                    - added nDataSize.
*
*********************************************************/
linklist *llnew(nCompare, nDataSize)
int nCompare;
int nDataSize;
{
   linklist *lpList;
   
   lpList = (linklist *)malloc(sizeof(linklist));
   if (lpList != NULL)
   {
      lpList->nDataSize = nDataSize;
      lpList->lpHead = NULL;
      lpList->lpTail = NULL;
      lpList->lpCurrent = NULL;
      lpList->lpfnCompare = calloc(nCompare,sizeof(int (*)()));
   }
   return(lpList);
}

/*********************************************************
* FUNCTION: 
*  lladdfind
*
* PROTOTYPE:
*  linklist *lladdfind(int (* )(), int);
*
* PARAMETERS:
*  int (* )(void *, void *)  compare function used for linklist element.
*  int                       index of compare function.
*
* RETURN:
*  linklist *  pointer to the new linklist.
*
* DESCRIPTION:
*  creates a link list.
*
* NOTES:
*  compare function should return 1 when found, 0 if not.
*
* UPDATES: 01/20/94  - added support for multiple compare 
*                      functions.
*                    - changed search to compare.
*
*********************************************************/
void lladdfind(lpListName,lpfnCompareFunc, nCompare)
linklist *lpListName;
int       (* lpfnCompareFunc)(void *, void *);
int       nCompare;
{
      lpListName->lpfnCompare[nCompare] = lpfnCompareFunc;
}


/*********************************************************
* FUNCTION:
*  llfree.
*
* PROTOTYPE:
*  int llfree(linklist *);
*
* PARAMETERS:
*  linklist *lpListName   link list header.
*
* RETURN:
*  LLNOERROR.
*
* DESCRIPTION:
*  free a link list, removing all its elements.
*
* NOTES:
*
* UPDATES: 01/20/94 - added LLNOERROR on return.
*
*********************************************************/
int llfree(lpListName)
linklist *lpListName;
{
   if (lpListName)
   {
      while(lpListName->lpHead != NULL)
      {
         lpListName->lpCurrent = (linkelement *)lpListName->lpHead->lpNext;
         free(lpListName->lpHead->lpData);
         free(lpListName->lpHead);
         lpListName->lpHead = lpListName->lpCurrent;
      }

      free(lpListName->lpfnCompare);
      free(lpListName);
   }
   return(LLNOERROR);
}

/*********************************************************
* FUNCTION:
*  lladd.
*
* PROTOTYPE:
*  int lladd(linklist *, void *, int);
*
* PARAMETERS:
*  linklist *lpListName  Link list definition.
*  void     *lpValue     Value to be inserted.
*  int      nLoc         before or after current 
*                          LLPREV before.
*                          LLNEXT after.
*
* RETURN:
*  int   LLNOERROR.
*        LLGENERROR error inserting, out of heap.
*        LLINVPARA  invalid nLoc.
*
* DESCRIPTION:
*  insert a new element in link list at current.
*  can insert before or after current.
*
* NOTES:
*
* UPDATES:  
*           05/10/94 - fixed bug in insert.
*           01/20/94 - added support for adding at current
*                      rather than tail. can specify 
*                      direction.
*                    - added support for using defined error.
*                    - removed nSize parameter.
*
*********************************************************/
int lladd(lpListName, lpValue, nLoc)
linklist *lpListName;
void     *lpValue;
int      nLoc;
{
   linkelement *lpElement;
   void        *lpData;
   int         nError = LLNOERROR; 

   lpElement = (linkelement *)malloc(sizeof(linkelement));
   lpData = malloc(lpListName->nDataSize);
   if ((lpElement != NULL) && (lpData != NULL))
   {
      memcpy(lpData,lpValue,lpListName->nDataSize);
      lpElement->lpData = lpData;

      if (lpListName->lpHead == NULL)
      {
         lpElement->lpNext = NULL;
         lpElement->lpPrevious = NULL;
         lpListName->lpHead = lpElement;
         lpListName->lpTail = lpElement;
         lpListName->lpCurrent = lpElement;
      }
      else
      {
         switch(nLoc)
         {
            case LLNEXT:
               lpElement->lpNext = lpListName->lpCurrent->lpNext;
               lpElement->lpPrevious = lpListName->lpCurrent;
               lpListName->lpCurrent->lpNext = lpElement;
               if (lpElement->lpNext == NULL)
                  lpListName->lpTail = lpElement;
          else
       lpElement->lpNext->lpPrevious = lpElement;
               lpListName->lpCurrent = lpElement;
               break;

            case LLPREV:
               lpElement->lpNext = lpListName->lpCurrent;
               lpElement->lpPrevious = lpListName->lpCurrent->lpPrevious;
               lpListName->lpCurrent->lpPrevious = lpElement;
               if (lpElement->lpPrevious == NULL)
                  lpListName->lpHead = lpElement;
          else
        lpElement->lpPrevious->lpNext = lpElement;
               lpListName->lpCurrent = lpElement;
               break;

            default:
               nError = LLINVPARA;
               break;
         }
      }
   }
   else
      nError = LLGENERROR;

   return(nError);
}


/*********************************************************
* FUNCTION:
*  lldelete.
*
* PROTOTYPE:
*  int lldelete(linklist *, linkelement *);
*
* PARAMETERS:
*  LinkLIst    *lpListName  Link list definition.
*  linkelement *lpElement   Element to be removed.
*
* RETURN:
*  int  LLNOERROR.
*       LLGENERROR error deleting.
*
* DESCRIPTION:
*  remove an element from the link list.
*
* NOTES:
*  The current pointer is updated in this fation
*  if deleting the head, points to the new head.
*  if deleting the tail, points to the new tail.
*  if deleting a node anywhere else in link,
*  points to the one link preceding the deleted one.
*
* UPDATES:  01/20/94 - fixed problem with delete.
*                    - added standard error.
*
*********************************************************/
int lldelete(lpListName, lpElement)
linklist    *lpListName;
linkelement *lpElement;
{
   int nNumber = 0;

   if ((nNumber = llnumber(lpListName)) == 0)
      return(LLGENERROR);

   if (nNumber == 1)
   {
      /* delete last element */
      lpListName->lpHead = NULL;
      lpListName->lpTail = NULL;
      lpListName->lpCurrent = NULL;
   }  
   else if (lpElement == lpListName->lpTail)
   {
      /* move tail back by one */
      lpListName->lpTail = (linkelement *)lpElement->lpPrevious;
      lpListName->lpCurrent = lpListName->lpTail;
      lpListName->lpCurrent->lpNext = NULL;
   }
   else if (lpElement == lpListName->lpHead)
   {
      /* move head up by one */
      lpListName->lpHead = (linkelement *)lpElement->lpNext;
      lpListName->lpCurrent = lpListName->lpHead;
      lpListName->lpCurrent->lpPrevious = NULL;
   }
   else
   {
      /* delete element anywhere else in list */
      lpListName->lpCurrent = (linkelement *)lpElement->lpNext;
      lpListName->lpCurrent->lpPrevious = lpElement->lpPrevious;
      lpListName->lpCurrent = (linkelement *)lpElement->lpPrevious;
      lpListName->lpCurrent->lpNext = lpElement->lpNext;
   }

   if (lpElement->lpData)
      free(lpElement->lpData);

   free(lpElement);
   return(LLNOERROR);
}

/*********************************************************
* FUNCTION:
*  llnumber.
*
* PROTOTYPE:
*  int llnumber(linklist *);
*
* PARAMETERS:
*  linklist    *lpListName   Link list definition.
*
* RETURN:
*  int number of elements in link list.
*
* DESCRIPTION:
*  return the number of elements in the link list.
*
* NOTES:
*
* UPDATES:
*
*********************************************************/
int llnumber(lpListName)
linklist *lpListName;
{
   int nNum = 0;
   linkelement *lpList;

   if (lpListName)
   {
      lpList = lpListName->lpHead;

      while(lpList != NULL)
      {
         nNum ++;
         lpList = (linkelement *)lpList->lpNext;
      }
   }
   
   return(nNum);
}


/*********************************************************
* FUNCTION:
*  llget.
*
* PROTOTYPE:
*  void *llget(linkelement *);
*
* PARAMETERS:
*  linkelement *lpElement     Element to retrieve the value from.
*
* RETURN:
*  void * pointer to data.
*         NULL if error.
*
* DESCRIPTION:
*  extract the data from a link element.
*
* NOTES:
*
* UPDATES:
*
*********************************************************/
void *llget(lpElement)
linkelement *lpElement;
{
   if (lpElement != NULL)
      return(lpElement->lpData);

   return(NULL);
}

/********************* END OF PRIMITIVES ****************/  

/*********************************************************
* FUNCTION:
*  llgetnext.
*
* PROTOTYPE:
*  void *llgetnext(linklist *);
*
* PARAMETERS:
*  linklist    *lpListName Link list definition.
*
* RETURN:
*  void * pointer to value.
*         return NULL if error.
*
* DESCRIPTION:
*  extract the value fallowing the current link in list.
*  the current link is moved forward by one.
*
* NOTES:
*
* UPDATES:
*
*********************************************************/
void *llgetnext(lpListName)
linklist *lpListName;
{
   if (lpListName && (lpListName->lpCurrent != lpListName->lpTail))
   {
      lpListName->lpCurrent = (linkelement *)lpListName->lpCurrent->lpNext;
      return(llget(lpListName->lpCurrent));
   }
   
   return(NULL);
}
   
/*********************************************************
* FUNCTION:
*  llgetprev.
*
* PROTOTYPE:
*  void *llgetprev(linklist *);
*
* PARAMETERS:
*  linklist    *lpListName Link list definition.
*
* RETURN:
*  void * pointer to value.
*         return NULL if error.
*
* DESCRIPTION:
*  extract the value preceding the current link in list.
*  The current link is moved backward by one.
*
* NOTES:
*
* UPDATES:
*
*********************************************************/
void *llgetprev(lpListName)
linklist *lpListName;
{
   if (lpListName->lpCurrent != lpListName->lpHead)
   {
      lpListName->lpCurrent = (linkelement *)lpListName->lpCurrent->lpPrevious;
      return(llget(lpListName->lpCurrent));
   }
   
   return(NULL);
}
   


/*********************************************************
* FUNCTION:
*  llfind.
*
* PROTOTYPE:
*  void *llfind(linklist *, void *, int);
*
* PARAMETERS:
*  linklist *lpListName   list name variable, link list.
*  void     *lpValue      value to compare for.
*  int       nCompare     compare function to use.
*
* RETURN:
*  void *   pointer to data found.
*
* DESCRIPTION:
*  compare a link list for a specific element containg data.
*
* NOTES:
*  uses the compare function associated with link list.
*  Comapre function should return LLFOUND when found, 
*  LLNOTFOUND if not.
*  The compare starts at the current linklist element. 
*  To compare from begining use llsetbeg().
*  Thse compare function only does a compare = or !=
*
* UPDATES: 01/20/94  - added support for multiple compare 
*                    - changed search to compare.
*                      functions.
*                    - added LLFOUND & LLNOTFOUND.
*
*********************************************************/
void *llfind(lpListName,lpValue, nCompare)
linklist *lpListName;
void     *lpValue;
int       nCompare;
{
   if (!llnumber(lpListName))
      return(NULL);

   if (lpListName->lpfnCompare) 
   {
      while (!(*lpListName->lpfnCompare[nCompare])(lpListName->lpCurrent->lpData, lpValue))
      {
         if (lpListName->lpCurrent == lpListName->lpTail)
            return(NULL);

         lpListName->lpCurrent = (linkelement *)lpListName->lpCurrent->lpNext;

         if (!lpListName->lpCurrent)
            return(NULL);
      }
   }
   return(lpListName->lpCurrent->lpData);
}

/*********************** DEBUG OPTION *******************/

/*********************************************************
* FUNCTION:
*  lldump.
*
* PROTOTYPE:
*  int lldump(linklist *);
*
* PARAMETERS:
*  linklist *lpListName   link list header.
*
* RETURN:
*  LLNOERROR.
*
* DESCRIPTION:
*  dump a link list.
*
* NOTES:
*
* UPDATES: 01/20/94 - created for debug purpose.
*                   - added LLGENERROR.
*
*********************************************************/
int lldump(lpListName)
linklist *lpListName;
{
   linkelement *lpLink;

   if (lpListName)
   {
      fprintf(stdout,"Head %x Tail %x Current %x\n",lpListName->lpHead,
                                                    lpListName->lpTail,
                                                    lpListName->lpCurrent);
      lpLink = lpListName->lpHead;
      while(lpLink != NULL)
      {
         fprintf(stdout," Link %x Next %x Prev %x Data %x\n",lpLink,
                                                             lpLink->lpNext,
                                                             lpLink->lpPrevious,
                                                             lpLink->lpData);
         lpLink = lpLink->lpNext;
      }
   }

   return (LLNOERROR);
}

