/**********************************************************
* FILE QUEUE.C version 1.00  - 05/10/94 -
*
* DESCRIPTION:
*  contains queue's functions.
*
* NOTES:
*
* Copyright 1993 by EZ Software 
*
* Permission to use, copy, modify, and distribute this software and its
* documentation for any purpose and without fee is hereby granted, provided 
* that the above copyright notice appear in all copies and that both that 
* copyright notice and this permission notice appear in supporting 
* documentation, and that the name of EZ Software 
* not be used in advertising or publicity pertaining to distribution of the 
* software without specific written prior permission. 
* EZ Software makes no representations about the 
* suitability of this software for any purpose.  It is provided "as is"
* without express or implied warranty.
*
* EZ Software DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
* SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, 
* IN NO EVENT SHALL EZ Software BE LIABLE FOR ANY SPECIAL,
* INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES RESULTING FROM LOSS OF USE,
* DATA OR PROFITS, WHETHER IN AN ACTION CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
* ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS 
* SOFTWARE.
*
*
*  Original Author: Laurent Chavey, EZ Software 
*           (chavey@cis.udel.edu)
*            june, 1993
*
**********************************************************/

#include "queue.h"

/********************* PRIMITIVES ***********************/

/*********************************************************
* FUNCTION: 
*  qnew
*
* PROTOTYPE:
*  queue *qnew(int );
*
* PARAMETERS:
*  int nDataSize   size in byte of the data element.
*
* RETURN:
*  queue *  pointer to the new queue.
*
* DESCRIPTION:
*  creates an empty queue.
*
* UPDATES: 
*
*********************************************************/
queue *qnew(nDataSize)
int nDataSize;
{
   queue *lpQueue;
   
   lpQueue = (queue *)malloc(sizeof(queue));
   if (lpQueue != NULL)
   {
      lpQueue->nDataSize = nDataSize;
      lpQueue->lpFront = NULL;
      lpQueue->lpBack = NULL;
   }

   return(lpQueue);
}

/*********************************************************
* FUNCTION:
*  qfree.
*
* PROTOTYPE:
*  int qfree(queue *);
*
* PARAMETERS:
*  queue *lpQueueName   queue header.
*
* RETURN:
*  QNOERROR.
*
* DESCRIPTION:
*  free a queue, removing all of its elements.
*
* NOTES:
*
* UPDATES: 
*
*********************************************************/
int qfree(lpQueueName)
queue *lpQueueName;
{
   queueelement *lpCurrent;

   while(lpQueueName->lpFront != NULL)
   {
      lpCurrent = (queueelement *)lpQueueName->lpFront;
      lpQueueName->lpFront = (queueelement *)lpQueueName->lpFront->lpNext;
      free(lpCurrent->lpData);
      free(lpCurrent);
   }
   free(lpQueueName);

   return(QNOERROR);
}

/*********************************************************
* FUNCTION:
*  qinsert.
*
* PROTOTYPE:
*  int qinsert(queue *, void *);
*
* PARAMETERS:
*  queue    *lpQueueName  queue definition.
*  void     *lpValue      Value to be inserted.
*
* RETURN:
*  int   QNOERROR.
*        QGENERROR error inserting, out of heap.
*
* DESCRIPTION:
*  add a new element to the back of the queue.
*
* NOTES:
*
* UPDATES:  
*
*********************************************************/
int qinsert(lpQueueName, lpValue)
queue *lpQueueName;
void  *lpValue;
{
   queueelement *lpElement;
   void         *lpData;
   int          nError = QNOERROR; 

   lpElement = (queueelement *)malloc(sizeof(queueelement));
   lpData = malloc(lpQueueName->nDataSize);
   if ((lpElement != NULL) && (lpData != NULL))
   {
      memcpy(lpData,lpValue,lpQueueName->nDataSize);
      lpElement->lpData = lpData;
      lpElement->lpNext = NULL;

      if (!lpQueueName->lpBack)
         lpQueueName->lpFront = lpElement;
      else
         lpQueueName->lpBack->lpNext = (struct queueelement*)lpElement;

      lpQueueName->lpBack = lpElement;
   }
   else
      nError = QGENERROR;

   return(nError);
}


/*********************************************************
* FUNCTION:
*  qremove.
*
* PROTOTYPE:
*  int qremove(queue *, queueelement *);
*
* PARAMETERS:
*  queue *lpQueeuName   queue definition.
*  void  *lpData        data returned.
*
* RETURN:
*  int  QNOERROR.
*       QEMPTY    queue is empty.
*
* DESCRIPTION:
*  remove the data from the front of the queue.
*
* NOTES:
*
* UPDATES:  
*
*********************************************************/
int qremove(lpQueueName, lpData)
queue  *lpQueueName;
void   *lpData;
{
   queueelement *lpCurrent;
   int          nError = QEMPTY;

   if (lpQueueName->lpFront)
   {
      /* get front element anrd remove it from queue */
      lpCurrent = lpQueueName->lpFront;
      lpQueueName->lpFront = (queueelement *)lpCurrent->lpNext;
      memcpy(lpData,lpCurrent->lpData,lpQueueName->nDataSize);
      free(lpCurrent->lpData);
      free(lpCurrent);
      if (!lpQueueName->lpFront)
         lpQueueName->lpBack = lpQueueName->lpFront;

      nError = QNOERROR;
   }
   return(nError);
}

/*********************************************************
* FUNCTION:
*  qnumber.
*
* PROTOTYPE:
*  int qnumber(queue *);
*
* PARAMETERS:
*  queue    *lpQueueName   Queue definition.
*
* RETURN:
*  int number of elements in the queue.
*
* DESCRIPTION:
*  return the number of elements in the queue.
*
* NOTES:
*
* UPDATES:
*
*********************************************************/
int qnumber(lpQueueName)
queue *lpQueueName;
{
   int nNum = 0;
   queueelement *lpQueueElement;

   lpQueueElement = lpQueueName->lpFront;

   while(lpQueueElement != NULL)
   {
      nNum ++;
      lpQueueElement = (queueelement *)lpQueueElement->lpNext;
   }
   
   return(nNum);
}


/*********************** DEBUG OPTION *******************/

/*********************************************************
* FUNCTION:
*  qdump.
*
* PROTOTYPE:
*  int qdump(queue *);
*
* PARAMETERS:
*  queue *lpQueueName   queue header.
*
* RETURN:
*  QNOERROR.
*
* DESCRIPTION:
*  dump a queue.
*
* NOTES:
*
* UPDATES: 
*
*********************************************************/
int qdump(lpQueueName)
queue *lpQueueName;
{
   queueelement *lpElement;

   if (lpQueueName)
   {
      fprintf(stdout,"back %x front %x \n",lpQueueName->lpBack,
                                           lpQueueName->lpFront);
      lpElement = lpQueueName->lpFront;
      while(lpElement != NULL)
      {
         fprintf(stdout,"Next %x Data %x\n",lpElement->lpNext,
                                            lpElement->lpData);
         lpElement = (queueelement *)lpElement->lpNext;
      }
   }

   return (QNOERROR);
}

